/*
 * Test dataspace operations.
 *
 * Copyright (C) 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <thread>

#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>

#include "dataspace_client.h"
#include "opener_client.h"
#include "opener_context_client.h"
#include "mapped_file_client.h"
#include "memory_utils.h"



/* Test parameters affected by capability limits. */

const unsigned long NUMBER_OF_FILES = 10;
const unsigned int START_LIMIT = 10;
const unsigned int ACTIVITY_ITERATIONS = 20;

/* Test parameters unaffected by any capability limits. */

const unsigned int MAP_PAGES = 20;
const unsigned int REGION_ITERATIONS = 10;



/* An activity opening and reading from a file. */

static long activity(l4_cap_idx_t context_ref, unsigned long fileid, unsigned int start_page)
{
  client_OpenerContext context(context_ref);
  unsigned long step = page(1);
  unsigned long sample = page(1);

  /* Allocate a buffer for sampling from the file. */

  char buf[sample + 1];

  /* Invoke the open method to receive the file reference. */

  unsigned long size;
  l4_cap_idx_t file_ref;

  long err = context.open(L4_FPAGE_RW, &size, &file_ref);

  if (err)
  {
    printf("Could not obtain file for %ld @ page %d: %s\n", fileid, start_page, l4sys_errtostr(err));
    return err;
  }

  client_MappedFile file(file_ref);

  /* Some memory to be mapped. */

  size_t start_pos, end_pos, data_end;

  err = file.mmap(page(start_page), page(MAP_PAGES), &start_pos, &end_pos, &data_end);

  if (err)
  {
    printf("Could not map file region for %ld @ page %d: %s\n", fileid, start_page, l4sys_errtostr(err));
    ipc_cap_free_um(file_ref);
    return err;
  }

  size = end_pos - start_pos;

  char *memory;

  err = ipc_attach_dataspace(file_ref, size, (void **) &memory);

  if (err)
  {
    printf("Could not map memory for %ld @ page %d: %s\n", fileid, start_page, l4sys_errtostr(err));
    ipc_cap_free_um(file_ref);
    return err;
  }

  if (data_end < size)
    size = data_end;

  /* Read the region a number of times. */

  for (unsigned int read_counter = 0; read_counter < REGION_ITERATIONS; read_counter++)
  {
    for (unsigned long offset = 0; offset < size; offset += step)
    {
      unsigned long remaining = size - offset;
      unsigned long sample_remaining = remaining < sample ? remaining : sample;

      strncpy(buf, (memory + offset), sample_remaining);
      buf[sample_remaining] = '\0';

      /* Test the data obtained. */

      unsigned long filepos = start_pos + offset;
      unsigned long _fileid = 0, _filepos = 0;
      char *sep = strchr(buf, ':');

      if (sep != NULL)
      {
        *sep = '\0'; sep++;
        _fileid = atol(buf); _filepos = atol(sep);
      }

      if ((fileid != _fileid) || (filepos != _filepos))
        printf("! %ld:%ld is not %ld:%ld\n", _fileid, _filepos, fileid, filepos);
    }
  }

  ipc_cap_free_um(file_ref);
  ipc_detach_dataspace(memory);

  return L4_EOK;
}



static long activity_iterate(l4_cap_idx_t context_ref, unsigned long fileid, unsigned int start_page)
{
  long err;

  /* Open the file, read pages, close the file, over and over. */

  for (unsigned int iteration = 0; iteration < ACTIVITY_ITERATIONS; iteration++)
  {
    err = activity(context_ref, fileid, start_page);
    if (err)
      break;
  }

  return err;
}



static long context_for_file(unsigned long fileid, l4_cap_idx_t *context_ref, char **filename)
{
  /* Obtain access to the filesystem. */

  l4_cap_idx_t server = l4re_env_get_cap("server");
  client_Opener opener(server);

  long err = opener.context(context_ref);

  if (err)
  {
    printf("Could not obtain context: %s\n", l4sys_errtostr(err));
    return err;
  }

  client_Dataspace context_ds(*context_ref);
  unsigned long size, flags;

  err = context_ds.info(&size, &flags);

  if (err)
  {
    printf("Could not obtain context info: %s\n", l4sys_errtostr(err));
    ipc_cap_free_um(*context_ref);
    return err;
  }

  /* Map context memory to write the filename. */

  err = ipc_attach_dataspace(*context_ref, size, (void **) filename);

  if (err)
  {
    printf("Could not map memory: %s\n", l4sys_errtostr(err));
    ipc_cap_free_um(*context_ref);
    return err;
  }

  /* Write the filename. */

  sprintf(*filename, "%ld", fileid);

  return L4_EOK;
}



static long activity_for_file(unsigned long fileid)
{
  std::thread *activities[START_LIMIT];

  l4_cap_idx_t context_ref;
  char *filename;
  int current = 0;

  long err = context_for_file(fileid, &context_ref, &filename);

  if (err)
    return err;

  for (unsigned int start_page = 0; start_page < START_LIMIT; start_page++)
    activities[current++] = new std::thread(activity_iterate, context_ref, fileid, start_page);

  /* Wait for the threads. */

  int limit = current;

  printf("Waiting for %d threads for %ld...\n", limit, fileid);

  for (current = 0; current < limit; current++)
  {
    activities[current]->join();
    printf("End: %ld @ %d\n", fileid, current);
  }

  /* Discard the context. */

  ipc_cap_free_um(context_ref);
  ipc_detach_dataspace(filename);

  printf("End: %ld\n", fileid);
  return L4_EOK;
}



int main(void)
{
  /* Introduce concurrency control. */

  ipc_cap_alloc_init();

  std::thread *activities[NUMBER_OF_FILES];

  unsigned long fileid;
  int current = 0;

  for (fileid = 0; fileid < NUMBER_OF_FILES; fileid++)
    activities[current++] = new std::thread(activity_for_file, fileid);

  /* Wait for the threads. */

  int limit = current;

  printf("Waiting for %d threads...\n", limit);

  for (current = 0; current < limit; current++)
  {
    activities[current]->join();
    printf("Ended: %d\n", current);
  }

  printf("Activities completed.\n");

  return 0;
}
